# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from dataclasses import dataclass
from typing import Union

import numpy.typing as ntp
import torch
import numpy as np

from adam.core.spatial_math import ArrayLike, ArrayLikeFactory, SpatialMath


@dataclass
class TorchLike(ArrayLike):
    """Class wrapping pyTorch types"""

    array: torch.Tensor

    def __post_init__(self):
        """Converts array to double precision"""
        if self.array.dtype != torch.float64:
            self.array = self.array.double()

    def __setitem__(self, idx, value: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides set item operator"""
        if type(self) is type(value):
            self.array[idx] = value.array.reshape(self.array[idx].shape)
        else:
            self.array[idx] = torch.tensor(value) if isinstance(value, float) else value

    def __getitem__(self, idx):
        """Overrides get item operator"""
        return TorchLike(self.array[idx])

    @property
    def shape(self):
        return self.array.shape

    def reshape(self, *args):
        return self.array.reshape(*args)

    @property
    def T(self) -> "TorchLike":
        """
        Returns:
            TorchLike: transpose of array
        """
        # check if self.array is a 0-D tensor

        if len(self.array.shape) == 0:
            return TorchLike(self.array)
        x = self.array
        return TorchLike(x.permute(*torch.arange(x.ndim - 1, -1, -1)))

    def __matmul__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides @ operator"""

        if type(self) is type(other):
            return TorchLike(self.array @ other.array)
        if isinstance(other, torch.Tensor):
            return TorchLike(self.array @ other)
        else:
            return TorchLike(self.array @ torch.tensor(other))

    def __rmatmul__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides @ operator"""
        if type(self) is type(other):
            return TorchLike(other.array @ self.array)
        else:
            return TorchLike(torch.tensor(other) @ self.array)

    def __mul__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides * operator"""
        if type(self) is type(other):
            return TorchLike(self.array * other.array)
        else:
            return TorchLike(self.array * other)

    def __rmul__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides * operator"""
        if type(self) is type(other):
            return TorchLike(other.array * self.array)
        else:
            return TorchLike(other * self.array)

    def __truediv__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides / operator"""
        if type(self) is type(other):
            return TorchLike(self.array / other.array)
        else:
            return TorchLike(self.array / other)

    def __add__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides + operator"""
        if type(self) is not type(other):
            return TorchLike(self.array.squeeze() + other.squeeze())
        return TorchLike(self.array.squeeze() + other.array.squeeze())

    def __radd__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides + operator"""
        if type(self) is not type(other):
            return TorchLike(self.array.squeeze() + other.squeeze())
        return TorchLike(self.array.squeeze() + other.array.squeeze())

    def __sub__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides - operator"""
        if type(self) is type(other):
            return TorchLike(self.array.squeeze() - other.array.squeeze())
        else:
            return TorchLike(self.array.squeeze() - other.squeeze())

    def __rsub__(self, other: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """Overrides - operator"""
        if type(self) is type(other):
            return TorchLike(other.array.squeeze() - self.array.squeeze())
        else:
            return TorchLike(other.squeeze() - self.array.squeeze())

    def __neg__(self) -> "TorchLike":
        """Overrides - operator"""
        return TorchLike(-self.array)


class TorchLikeFactory(ArrayLikeFactory):
    @staticmethod
    def zeros(*x: int) -> "TorchLike":
        """
        Returns:
            TorchLike: zero matrix of dimension *x
        """
        return TorchLike(torch.zeros(x))

    @staticmethod
    def eye(x: int) -> "TorchLike":
        """
        Args:
            x (int): dimension

        Returns:
            TorchLike: identity matrix of dimension x
        """
        return TorchLike(torch.eye(x))

    @staticmethod
    def array(x: ntp.ArrayLike) -> "TorchLike":
        """
        Returns:
            TorchLike: vector wrapping x
        """
        return TorchLike(torch.tensor(x))


class SpatialMath(SpatialMath):
    def __init__(self):
        super().__init__(TorchLikeFactory())

    @staticmethod
    def sin(x: ntp.ArrayLike) -> "TorchLike":
        """
        Args:
            x (ntp.ArrayLike): angle value

        Returns:
            TorchLike: sin value of x
        """
        if isinstance(x, float):
            x = torch.tensor(x)
        return TorchLike(torch.sin(x))

    @staticmethod
    def cos(x: ntp.ArrayLike) -> "TorchLike":
        """
        Args:
            x (ntp.ArrayLike): angle value

        Returns:
            TorchLike: cos value of x
        """
        # transform to torch tensor, if not already
        if isinstance(x, float):
            x = torch.tensor(x)
        return TorchLike(torch.cos(x))

    @staticmethod
    def outer(x: ntp.ArrayLike, y: ntp.ArrayLike) -> "TorchLike":
        """
        Args:
            x (ntp.ArrayLike): vector
            y (ntp.ArrayLike): vector

        Returns:
            TorchLike: outer product of x and y
        """
        return TorchLike(torch.outer(torch.tensor(x), torch.tensor(y)))

    @staticmethod
    def skew(x: Union["TorchLike", ntp.ArrayLike]) -> "TorchLike":
        """
        Args:
            x (Union[TorchLike, ntp.ArrayLike]): vector

        Returns:
            TorchLike: skew matrix from x
        """
        if not isinstance(x, TorchLike):
            return TorchLike(
                torch.tensor([[0, -x[2], x[1]], [x[2], 0, -x[0]], [-x[1], x[0], 0]])
            )
        x = x.array
        return TorchLike(
            torch.tensor([[0, -x[2], x[1]], [x[2], 0, -x[0]], [-x[1], x[0], 0]])
        )

    @staticmethod
    def vertcat(*x: ntp.ArrayLike) -> "TorchLike":
        """
        Returns:
            TorchLike: vertical concatenation of x
        """
        if isinstance(x[0], TorchLike):
            v = torch.vstack([x[i].array for i in range(len(x))])
        else:
            v = torch.tensor(x)
        return TorchLike(v)

    @staticmethod
    def horzcat(*x: ntp.ArrayLike) -> "TorchLike":
        """
        Returns:
            TorchLike: horizontal concatenation of x
        """
        if isinstance(x[0], TorchLike):
            v = torch.hstack([x[i].array for i in range(len(x))])
        else:
            v = torch.tensor(x)
        return TorchLike(v)



# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

from dataclasses import dataclass
from typing import Union

import jax.numpy as jnp
import numpy.typing as npt

from adam.core.spatial_math import ArrayLike, ArrayLikeFactory, SpatialMath
from adam.numpy import NumpyLike


@dataclass
class JaxLike(ArrayLike):
    """Wrapper class for Jax types"""

    array: jnp.array

    def __setitem__(self, idx, value: Union["JaxLike", npt.ArrayLike]):
        """Overrides set item operator"""
        if type(self) is type(value):
            self.array = self.array.at[idx].set(
                value.array.reshape(self.array[idx].shape)
            )
        else:
            self.array = self.array.at[idx].set(value)

    def __getitem__(self, idx) -> "JaxLike":
        """Overrides get item operator"""
        return JaxLike(self.array[idx])

    @property
    def shape(self):
        return self.array.shape

    def reshape(self, *args):
        return self.array.reshape(*args)

    @property
    def T(self) -> "JaxLike":
        """
        Returns:
            JaxLike: transpose of the array
        """
        return JaxLike(self.array.T)

    def __matmul__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides @ operator"""
        if type(other) in [JaxLike, NumpyLike]:
            return JaxLike(self.array @ other.array)
        else:
            return JaxLike(self.array @ jnp.array(other))

    def __rmatmul__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides @ operator"""
        if type(other) in [JaxLike, NumpyLike]:
            return JaxLike(other.array * self.array)
        else:
            return JaxLike(jnp.array(other) * self.array)

    def __mul__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides * operator"""
        if type(self) is not type(other):
            return JaxLike(self.array * other)
        return JaxLike(self.array * other.array)

    def __rmul__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides * operator"""
        if type(self) is not type(other):
            return JaxLike(other * self.array)
        return JaxLike(other.array * self.array)

    def __truediv__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides / operator"""
        if type(self) is not type(other):
            return JaxLike(self.array / other)
        return JaxLike(self.array / other.array)

    def __add__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides + operator"""
        if type(self) is not type(other):
            return JaxLike(self.array.squeeze() + other.squeeze())
        return JaxLike(self.array.squeeze() + other.array.squeeze())

    def __radd__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides + operator"""
        if type(self) is not type(other):
            return JaxLike(self.array.squeeze() + other.squeeze())
        return JaxLike(self.array.squeeze() + other.array.squeeze())

    def __sub__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides - operator"""

        if type(self) is not type(other):
            return JaxLike(self.array.squeeze() - other.squeeze())
        return JaxLike(self.array.squeeze() - other.array.squeeze())

    def __rsub__(self, other: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """Overrides - operator"""
        if type(self) is not type(other):
            return JaxLike(other.squeeze() - self.array.squeeze())
        return JaxLike(other.array.squeeze() - self.array.squeeze())

    def __neg__(self) -> "JaxLike":
        """Overrides - operator"""
        return JaxLike(-self.array)


class JaxLikeFactory(ArrayLikeFactory):
    @staticmethod
    def zeros(*x) -> "JaxLike":
        """
        Returns:
            JaxLike: Matrix of zeros of dim *x
        """
        return JaxLike(jnp.zeros(x))

    @staticmethod
    def eye(x) -> "JaxLike":
        """
        Returns:
            JaxLike: Identity matrix of dimension x
        """
        return JaxLike(jnp.eye(x))

    @staticmethod
    def array(x) -> "JaxLike":
        """
        Returns:
            JaxLike: Vector wrapping *x
        """
        return JaxLike(jnp.array(x))


class SpatialMath(SpatialMath):
    def __init__(self):
        super().__init__(JaxLikeFactory())

    @staticmethod
    def sin(x: npt.ArrayLike) -> "JaxLike":
        """
        Args:
            x (npt.ArrayLike): angle value

        Returns:
            JaxLike: sin of x
        """
        return JaxLike(jnp.sin(x))

    @staticmethod
    def cos(x: npt.ArrayLike) -> "JaxLike":
        """
        Args:
            x (npt.ArrayLike): angle value

        Returns:
            JaxLike: cos of x
        """
        return JaxLike(jnp.cos(x))

    @staticmethod
    def outer(x: npt.ArrayLike, y: npt.ArrayLike) -> "JaxLike":
        """
        Args:
            x (npt.ArrayLike): vector
            y (npt.ArrayLike): vector

        Returns:
            JaxLike: outer product between x and y
        """
        x = jnp.array(x)
        y = jnp.array(y)
        return JaxLike(jnp.outer(x, y))

    @staticmethod
    def skew(x: Union["JaxLike", npt.ArrayLike]) -> "JaxLike":
        """
        Args:
            x (Union[JaxLike, npt.ArrayLike]): vector

        Returns:
            JaxLike: the skew symmetric matrix from x
        """
        if not isinstance(x, JaxLike):
            return -jnp.cross(jnp.array(x), jnp.eye(3), axisa=0, axisb=0)
        x = x.array
        return JaxLike(-jnp.cross(jnp.array(x), jnp.eye(3), axisa=0, axisb=0))

    @staticmethod
    def vertcat(*x) -> "JaxLike":
        """
        Returns:
            JaxLike: Vertical concatenation of elements
        """
        if isinstance(x[0], JaxLike):
            v = jnp.vstack([x[i].array for i in range(len(x))])
        else:
            v = jnp.vstack([x[i] for i in range(len(x))])
        return JaxLike(v)

    @staticmethod
    def horzcat(*x) -> "JaxLike":
        """
        Returns:
            JaxLike: Horizontal concatenation of elements
        """
        if isinstance(x[0], JaxLike):
            v = jnp.hstack([x[i].array for i in range(len(x))])
        else:
            v = jnp.hstack([x[i] for i in range(len(x))])
        return JaxLike(v)



import pathlib
from typing import List

import urdf_parser_py.urdf
from adam.core.spatial_math import SpatialMath
from adam.model import ModelFactory, StdJoint, StdLink, Link, Joint
from adam.model.std_factories.std_model import urdf_remove_sensors_tags
from adam.parametric.model import ParametricJoint, ParametricLink


class URDFParametricModelFactory(ModelFactory):
    """This factory generates robot elements from urdf_parser_py parametrized w.r.t. link lengths and densities

    Args:
        ModelFactory: the Model factory
    """

    def __init__(
        self,
        path: str,
        math: SpatialMath,
        links_name_list: List,
        length_multiplier,
        densities,
    ):
        self.math = math
        if type(path) is not pathlib.Path:
            path = pathlib.Path(path)
        if not path.exists():
            raise FileExistsError(path)
        self.links_name_list = links_name_list

        # Read URDF, but before passing it to urdf_parser_py get rid of all sensor tags
        # sensor tags are valid elements of URDF (see ),
        # but they are ignored by urdf_parser_py, that complains every time it sees one.
        # As there is nothing to be fixed in the used models, and it is not useful
        # to have a useless and noisy warning, let's remove before hands all the sensor elements,
        # that anyhow are not parser by urdf_parser_py or adam
        # See https://github.com/ami-iit/ADAM/issues/59
        xml_file = open(path, "r")
        xml_string = xml_file.read()
        xml_file.close()
        xml_string_without_sensors_tags = urdf_remove_sensors_tags(xml_string)

        self.urdf_desc = urdf_parser_py.urdf.URDF.from_xml_string(
            xml_string_without_sensors_tags
        )
        self.name = self.urdf_desc.name
        self.length_multiplier = length_multiplier
        self.densities = densities

    def get_joints(self) -> List[Joint]:
        """
        Returns:
            List[Joint]: build the list of the joints
        """
        return [self.build_joint(j) for j in self.urdf_desc.joints]

    def get_links(self) -> List[Link]:
        """
        Returns:
            List[Link]: build the list of the links
        """
        return [
            self.build_link(l) for l in self.urdf_desc.links if l.inertial is not None
        ]

    def get_frames(self) -> List[StdLink]:
        """
        Returns:
            List[Link]: build the list of the links
        """
        return [self.build_link(l) for l in self.urdf_desc.links if l.inertial is None]

    def build_joint(self, joint: urdf_parser_py.urdf.Joint) -> Joint:
        """
        Args:
            joint (Joint): the urdf_parser_py joint

        Returns:
            StdJoint/ParametricJoint: our joint representation
        """
        if joint.parent in self.links_name_list:
            index_link = self.links_name_list.index(joint.parent)
            link_parent = self.get_element_by_name(joint.parent)
            parent_link_parametric = ParametricLink(
                link_parent,
                self.math,
                self.length_multiplier[index_link],
                self.densities[index_link],
            )
            return ParametricJoint(joint, self.math, parent_link_parametric)

        return StdJoint(joint, self.math)

    def build_link(self, link: urdf_parser_py.urdf.Link) -> Link:
        """
        Args:
            link (Link): the urdf_parser_py link

        Returns:
            StdLink/ParametricLink: our link representation
        """
        if link.name in self.links_name_list:
            index_link = self.links_name_list.index(link.name)
            return ParametricLink(
                link,
                self.math,
                self.length_multiplier[index_link],
                self.densities[index_link],
            )
        return StdLink(link, self.math)

    def get_element_by_name(self, link_name):
        """
        Args:
            link_name (Link): the link name

        Returns:
            Link: the urdf parser link object associated to the link name
        """
        link_list = [
            corresponding_link
            for corresponding_link in self.urdf_desc.links
            if corresponding_link.name == link_name
        ]
        if len(link_list) != 0:
            return link_list[0]
        else:
            return None



import pathlib
from typing import List
import xml.etree.ElementTree as ET

import urdf_parser_py.urdf

from adam.core.spatial_math import SpatialMath
from adam.model import ModelFactory, StdJoint, StdLink


def urdf_remove_sensors_tags(xml_string):
    # Parse the XML string
    root = ET.fromstring(xml_string)

    # Find and remove all tags named "sensor" that are child of
    # root node (i.e. robot)
    for sensors_tag in root.findall("sensor"):
        root.remove(sensors_tag)

    # Convert the modified XML back to a string
    modified_xml_string = ET.tostring(root)

    return modified_xml_string


class URDFModelFactory(ModelFactory):
    """This factory generates robot elements from urdf_parser_py

    Args:
        ModelFactory: the Model factory
    """

    def __init__(self, path: str, math: SpatialMath):
        self.math = math
        if type(path) is not pathlib.Path:
            path = pathlib.Path(path)
        if not path.exists():
            raise FileExistsError(path)

        # Read URDF, but before passing it to urdf_parser_py get rid of all sensor tags
        # sensor tags are valid elements of URDF (see ),
        # but they are ignored by urdf_parser_py, that complains every time it sees one.
        # As there is nothing to be fixed in the used models, and it is not useful
        # to have a useless and noisy warning, let's remove before hands all the sensor elements,
        # that anyhow are not parser by urdf_parser_py or adam
        # See https://github.com/ami-iit/ADAM/issues/59
        xml_file = open(path, "r")
        xml_string = xml_file.read()
        xml_file.close()
        xml_string_without_sensors_tags = urdf_remove_sensors_tags(xml_string)
        self.urdf_desc = urdf_parser_py.urdf.URDF.from_xml_string(
            xml_string_without_sensors_tags
        )
        self.name = self.urdf_desc.name

    def get_joints(self) -> List[StdJoint]:
        """
        Returns:
            List[StdJoint]: build the list of the joints
        """
        return [self.build_joint(j) for j in self.urdf_desc.joints]

    def get_links(self) -> List[StdLink]:
        """
        Returns:
            List[StdLink]: build the list of the links
        """
        return [
            self.build_link(l) for l in self.urdf_desc.links if l.inertial is not None
        ]

    def get_frames(self) -> List[StdLink]:
        """
        Returns:
            List[StdLink]: build the list of the links
        """
        return [self.build_link(l) for l in self.urdf_desc.links if l.inertial is None]

    def build_joint(self, joint: urdf_parser_py.urdf.Joint) -> StdJoint:
        """
        Args:
            joint (Joint): the urdf_parser_py joint

        Returns:
            StdJoint: our joint representation
        """
        return StdJoint(joint, self.math)

    def build_link(self, link: urdf_parser_py.urdf.Link) -> StdLink:
        """
        Args:
            link (Link): the urdf_parser_py link

        Returns:
            StdLink: our link representation
        """
        return StdLink(link, self.math)



import abc
import dataclasses
from typing import List

import numpy.typing as npt

from adam.core.spatial_math import SpatialMath


@dataclasses.dataclass
class Pose:
    """Pose class"""

    xyz: List
    rpy: List


@dataclasses.dataclass
class Inertia:
    """Inertia class"""

    ixx: npt.ArrayLike
    ixy: npt.ArrayLike
    ixz: npt.ArrayLike
    iyy: npt.ArrayLike
    iyz: npt.ArrayLike
    izz: npt.ArrayLike


@dataclasses.dataclass
class Limits:
    """Limits class"""

    lower: npt.ArrayLike
    upper: npt.ArrayLike
    effort: npt.ArrayLike
    velocity: npt.ArrayLike


@dataclasses.dataclass
class Joint(abc.ABC):
    """Base Joint class. You need to fill at least these fields"""

    math: SpatialMath
    name: str
    parent: str
    child: str
    type: str
    axis: List
    origin: Pose
    limit: Limits
    idx: int
    """
    Abstract base class for all joints.
    """

    @abc.abstractmethod
    def spatial_transform(self, q: npt.ArrayLike) -> npt.ArrayLike:
        """
        Args:
            q (npt.ArrayLike): joint motion

        Returns:
            npt.ArrayLike: spatial transform of the joint given q
        """
        pass

    @abc.abstractmethod
    def motion_subspace(self) -> npt.ArrayLike:
        """
        Returns:
            npt.ArrayLike: motion subspace of the joint
        """

    @abc.abstractmethod
    def homogeneous(self, q: npt.ArrayLike) -> npt.ArrayLike:
        """
        Args:
            q (npt.ArrayLike): joint value
        Returns:
            npt.ArrayLike: homogeneous transform given the joint value
        """
        pass


@dataclasses.dataclass
class Inertial:
    """Inertial description"""

    mass: npt.ArrayLike
    inertia = Inertia
    origin = Pose


@dataclasses.dataclass
class Link(abc.ABC):
    """Base Link class. You need to fill at least these fields"""

    math: SpatialMath
    name: str
    visuals: List
    inertial: Inertial
    collisions: List

    @abc.abstractmethod
    def spatial_inertia(self) -> npt.ArrayLike:
        """
        Returns:
            npt.ArrayLike: the 6x6 inertia matrix expressed at
                           the origin of the link (with rotation)
        """
        pass

    @abc.abstractmethod
    def homogeneous(self) -> npt.ArrayLike:
        """
        Returns:
            npt.ArrayLike: the homogeneous transform of the link
        """
        pass


@dataclasses.dataclass
class ModelFactory(abc.ABC):
    """The abstract class of the model factory.

    The model factory is responsible for creating the model.

    You need to implement all the methods in your concrete implementation
    """

    math: SpatialMath
    name: str

    @abc.abstractmethod
    def build_link(self) -> Link:
        """build the single link
        Returns:
            Link
        """
        pass

    @abc.abstractmethod
    def build_joint(self) -> Joint:
        """build the single joint

        Returns:
            Joint
        """
        pass

    @abc.abstractmethod
    def get_links(self) -> List[Link]:
        """
        Returns:
            List[Link]: the list of the link
        """
        pass

    @abc.abstractmethod
    def get_frames(self) -> List[Link]:
        """
        Returns:
            List[Link]: the list of the frames
        """
        pass

    @abc.abstractmethod
    def get_joints(self) -> List[Joint]:
        """
        Returns:
            List[Joint]: the list of the joints
        """
        pass



# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import logging
import casadi as cs
import idyntree.swig as idyntree
import numpy as np
import pytest

import adam.numpy
from adam.parametric.casadi import KinDynComputationsParametric
from adam.parametric.model.parametric_factories.parametric_model import (
    URDFParametricModelFactory,
)
from adam.model.conversions.idyntree import to_idyntree_model
from adam.core.constants import Representations

from adam.geometry import utils
import tempfile
from git import Repo

# Getting stickbot urdf file
temp_dir = tempfile.TemporaryDirectory()
git_url = "https://github.com/icub-tech-iit/ergocub-gazebo-simulations.git"
Repo.clone_from(git_url, temp_dir.name)
model_path = temp_dir.name + "/models/stickBot/model.urdf"

joints_name_list = [
    "torso_pitch",
    "torso_roll",
    "torso_yaw",
    "l_shoulder_pitch",
    "l_shoulder_roll",
    "l_shoulder_yaw",
    "l_elbow",
    "r_shoulder_pitch",
    "r_shoulder_roll",
    "r_shoulder_yaw",
    "r_elbow",
    "l_hip_pitch",
    "l_hip_roll",
    "l_hip_yaw",
    "l_knee",
    "l_ankle_pitch",
    "l_ankle_roll",
    "r_hip_pitch",
    "r_hip_roll",
    "r_hip_yaw",
    "r_knee",
    "r_ankle_pitch",
    "r_ankle_roll",
]


logging.basicConfig(level=logging.DEBUG)
logging.debug("Showing the robot tree.")


root_link = "root_link"

link_name_list = ["chest"]
comp_w_hardware = KinDynComputationsParametric(
    model_path, joints_name_list, link_name_list, root_link
)
comp_w_hardware.set_frame_velocity_representation(Representations.MIXED_REPRESENTATION)

original_density = [
    628.0724496264945
]  # This is the original density value associated to the chest link, computed as mass/volume
original_length = np.ones(len(link_name_list))

# TODO: the following two commands should be moved to a proper function/method
factory = URDFParametricModelFactory(
    path=model_path,
    math=adam.numpy.numpy_like.SpatialMath(),
    links_name_list=link_name_list,
    length_multiplier=original_length,
    densities=original_density,
)
model = adam.model.Model.build(factory=factory, joints_name_list=joints_name_list)


kinDyn = idyntree.KinDynComputations()
kinDyn.loadRobotModel(to_idyntree_model(model))
kinDyn.setFloatingBase(root_link)
kinDyn.setFrameVelocityRepresentation(idyntree.MIXED_REPRESENTATION)
n_dofs = len(joints_name_list)


# base pose quantities
xyz = (np.random.rand(3) - 0.5) * 5
rpy = (np.random.rand(3) - 0.5) * 5
base_vel = (np.random.rand(6) - 0.5) * 5
# joints quantities
joints_val = (np.random.rand(n_dofs) - 0.5) * 5
joints_dot_val = (np.random.rand(n_dofs) - 0.5) * 5

H_b = utils.H_from_Pos_RPY(xyz, rpy)
vb_ = base_vel
s_ = joints_val
s_dot_ = joints_dot_val

g = np.array([0, 0, -9.80665])
kinDyn.setRobotState(H_b, joints_val, base_vel, joints_dot_val, g)


def test_mass_matrix():
    mass_mx = idyntree.MatrixDynSize()
    kinDyn.getFreeFloatingMassMatrix(mass_mx)
    mass_mxNumpy = mass_mx.toNumPy()
    M_with_hardware = comp_w_hardware.mass_matrix_fun()
    mass_test_hardware = cs.DM(
        M_with_hardware(H_b, s_, original_length, original_density)
    )
    assert mass_mxNumpy - mass_test_hardware == pytest.approx(0.0, abs=1e-5)


def test_CMM():
    cmm_idyntree = idyntree.MatrixDynSize()
    kinDyn.getCentroidalTotalMomentumJacobian(cmm_idyntree)
    cmm_idyntreeNumpy = cmm_idyntree.toNumPy()
    Jcm_with_hardware = comp_w_hardware.centroidal_momentum_matrix_fun()
    Jcm_test_hardware = cs.DM(
        Jcm_with_hardware(H_b, s_, original_length, original_density)
    )
    assert cmm_idyntreeNumpy - Jcm_test_hardware == pytest.approx(0.0, abs=1e-5)


def test_CoM_pos():
    CoM_iDynTree = kinDyn.getCenterOfMassPosition().toNumPy()
    com_with_hardware_f = comp_w_hardware.CoM_position_fun()
    CoM_hardware = cs.DM(
        com_with_hardware_f(H_b, s_, original_length, original_density)
    )
    assert CoM_iDynTree - CoM_hardware == pytest.approx(0.0, abs=1e-5)


def test_total_mass():
    mass = kinDyn.model().getTotalMass()
    mass_hardware_fun = comp_w_hardware.get_total_mass()
    mass_hardware = cs.DM(mass_hardware_fun(original_length, original_density))
    assert mass - mass_hardware == pytest.approx(0.0, abs=1e-5)


def test_jacobian():
    iDyntreeJ_ = idyntree.MatrixDynSize(6, n_dofs + 6)
    kinDyn.getFrameFreeFloatingJacobian("l_sole", iDyntreeJ_)
    iDynNumpyJ_ = iDyntreeJ_.toNumPy()
    J_tot_with_hardware = comp_w_hardware.jacobian_fun("l_sole")
    J_test_hardware = cs.DM(
        J_tot_with_hardware(H_b, s_, original_length, original_density)
    )
    assert iDynNumpyJ_ - J_test_hardware == pytest.approx(0.0, abs=1e-5)


def test_jacobian_non_actuated():
    iDyntreeJ_ = idyntree.MatrixDynSize(6, n_dofs + 6)
    kinDyn.getFrameFreeFloatingJacobian("head", iDyntreeJ_)
    iDynNumpyJ_ = iDyntreeJ_.toNumPy()
    J_tot_with_hardware = comp_w_hardware.jacobian_fun("head")
    J_tot_with_hardware_test = cs.DM(
        J_tot_with_hardware(H_b, s_, original_length, original_density)
    )
    assert iDynNumpyJ_ - J_tot_with_hardware_test == pytest.approx(0.0, abs=1e-5)


def test_jacobian_dot():
    Jdotnu = kinDyn.getFrameBiasAcc("l_sole")
    Jdot_nu = Jdotnu.toNumPy()
    J_dot_hardware = comp_w_hardware.jacobian_dot_fun("l_sole")
    J_dot_nu_test2 = cs.DM(
        J_dot_hardware(
            H_b, joints_val, base_vel, joints_dot_val, original_length, original_density
        )
        @ np.concatenate((base_vel, joints_dot_val))
    )
    assert Jdot_nu - J_dot_nu_test2 == pytest.approx(0.0, abs=1e-5)


def test_fk():
    H_idyntree = kinDyn.getWorldTransform("l_sole")
    p_idy2np = H_idyntree.getPosition().toNumPy()
    R_idy2np = H_idyntree.getRotation().toNumPy()
    T_with_hardware = comp_w_hardware.forward_kinematics_fun("l_sole")
    H_with_hardware_test = cs.DM(
        T_with_hardware(H_b, s_, original_length, original_density)
    )
    assert H_with_hardware_test[:3, :3] - R_idy2np == pytest.approx(0.0, abs=1e-5)
    assert H_with_hardware_test[:3, 3] - p_idy2np == pytest.approx(0.0, abs=1e-5)


def test_fk_non_actuated():
    H_idyntree = kinDyn.getWorldTransform("head")
    p_idy2np = H_idyntree.getPosition().toNumPy()
    R_idy2np = H_idyntree.getRotation().toNumPy()
    T_with_hardware = comp_w_hardware.forward_kinematics_fun("head")
    H_with_hardware_test = cs.DM(
        T_with_hardware(H_b, s_, original_length, original_density)
    )
    assert H_with_hardware_test[:3, :3] - R_idy2np == pytest.approx(0.0, abs=1e-5)
    assert H_with_hardware_test[:3, 3] - p_idy2np == pytest.approx(0.0, abs=1e-5)


def test_bias_force():
    h_iDyn = idyntree.FreeFloatingGeneralizedTorques(kinDyn.model())
    assert kinDyn.generalizedBiasForces(h_iDyn)
    h_iDyn_np = np.concatenate(
        (h_iDyn.baseWrench().toNumPy(), h_iDyn.jointTorques().toNumPy())
    )

    h_with_hardware = comp_w_hardware.bias_force_fun()
    h_with_hardware_test = cs.DM(
        h_with_hardware(H_b, s_, vb_, s_dot_, original_length, original_density)
    )
    assert h_with_hardware_test - h_iDyn_np == pytest.approx(0.0, abs=1e-4)


def test_coriolis_term():
    g0 = idyntree.Vector3()
    g0.zero()
    kinDyn.setRobotState(H_b, joints_val, base_vel, joints_dot_val, g0)
    C_iDyn = idyntree.FreeFloatingGeneralizedTorques(kinDyn.model())
    assert kinDyn.generalizedBiasForces(C_iDyn)
    C_iDyn_np = np.concatenate(
        (C_iDyn.baseWrench().toNumPy(), C_iDyn.jointTorques().toNumPy())
    )
    C_with_hardware = comp_w_hardware.coriolis_term_fun()
    C_with_hardware_test = cs.DM(
        C_with_hardware(H_b, s_, vb_, s_dot_, original_length, original_density)
    )
    assert C_with_hardware_test - C_iDyn_np == pytest.approx(0.0, abs=1e-4)


def test_gravity_term():
    base_vel_0 = np.zeros(6)
    joints_dot_val_0 = np.zeros(n_dofs)
    kinDyn.setRobotState(H_b, joints_val, base_vel_0, joints_dot_val_0, g)
    G_iDyn = idyntree.FreeFloatingGeneralizedTorques(kinDyn.model())
    assert kinDyn.generalizedBiasForces(G_iDyn)
    G_iDyn_np = np.concatenate(
        (G_iDyn.baseWrench().toNumPy(), G_iDyn.jointTorques().toNumPy())
    )
    G_with_hardware = comp_w_hardware.gravity_term_fun()
    G_with_hardware_test = G_with_hardware(H_b, s_, original_length, original_density)
    assert G_with_hardware_test - G_iDyn_np == pytest.approx(0.0, abs=1e-4)



# Copyright (C) 2021 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import casadi as cs
import numpy as np
from typing import Union

from adam.casadi.casadi_like import SpatialMath
from adam.core import RBDAlgorithms
from adam.core.constants import Representations
from adam.model import Model, URDFModelFactory


class KinDynComputations:
    """This is a small class that retrieves robot quantities represented in a symbolic fashion using CasADi
    in mixed representation, for Floating Base systems - as humanoid robots.
    """

    def __init__(
        self,
        urdfstring: str,
        joints_name_list: list = None,
        root_link: str = "root_link",
        gravity: np.array = np.array([0.0, 0.0, -9.80665, 0.0, 0.0, 0.0]),
        f_opts: dict = dict(jit=False, jit_options=dict(flags="-Ofast")),
    ) -> None:
        """
        Args:
            urdfstring (str): path of the urdf
            joints_name_list (list): list of the actuated joints
            root_link (str, optional): the first link. Defaults to 'root_link'.
        """
        math = SpatialMath()
        factory = URDFModelFactory(path=urdfstring, math=math)
        model = Model.build(factory=factory, joints_name_list=joints_name_list)
        self.rbdalgos = RBDAlgorithms(model=model, math=math)
        self.NDoF = self.rbdalgos.NDoF
        self.g = gravity
        self.f_opts = f_opts

    def set_frame_velocity_representation(
        self, representation: Representations
    ) -> None:
        """Sets the representation of the velocity of the frames

        Args:
            representation (Representations): The representation of the velocity
        """
        self.rbdalgos.set_frame_velocity_representation(representation)

    def mass_matrix_fun(self) -> cs.Function:
        """Returns the Mass Matrix functions computed the CRBA

        Returns:
            M (casADi function): Mass Matrix
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        [M, _] = self.rbdalgos.crba(base_transform, joint_positions)
        return cs.Function(
            "M", [base_transform, joint_positions], [M.array], self.f_opts
        )

    def centroidal_momentum_matrix_fun(self) -> cs.Function:
        """Returns the Centroidal Momentum Matrix functions computed the CRBA

        Returns:
            Jcc (casADi function): Centroidal Momentum matrix
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        [_, Jcm] = self.rbdalgos.crba(base_transform, joint_positions)
        return cs.Function(
            "Jcm", [base_transform, joint_positions], [Jcm.array], self.f_opts
        )

    def forward_kinematics_fun(self, frame: str) -> cs.Function:
        """Computes the forward kinematics relative to the specified frame

        Args:
            frame (str): The frame to which the fk will be computed

        Returns:
            H (casADi function): The fk represented as Homogenous transformation matrix
        """
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_transform = cs.SX.sym("H", 4, 4)
        H = self.rbdalgos.forward_kinematics(frame, base_transform, joint_positions)
        return cs.Function(
            "H", [base_transform, joint_positions], [H.array], self.f_opts
        )

    def jacobian_fun(self, frame: str) -> cs.Function:
        """Returns the Jacobian relative to the specified frame

        Args:
            frame (str): The frame to which the jacobian will be computed

        Returns:
            J_tot (casADi function): The Jacobian relative to the frame
        """
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_transform = cs.SX.sym("H", 4, 4)
        J_tot = self.rbdalgos.jacobian(frame, base_transform, joint_positions)
        return cs.Function(
            "J_tot", [base_transform, joint_positions], [J_tot.array], self.f_opts
        )

    def relative_jacobian_fun(self, frame: str) -> cs.Function:
        """Returns the Jacobian between the root link and a specified frame frames

        Args:
            frame (str): The tip of the chain

        Returns:
            J (casADi function): The Jacobian between the root and the frame
        """
        joint_positions = cs.SX.sym("s", self.NDoF)
        J = self.rbdalgos.relative_jacobian(frame, joint_positions)
        return cs.Function("J", [joint_positions], [J.array], self.f_opts)

    def jacobian_dot_fun(self, frame: str) -> cs.Function:
        """Returns the Jacobian derivative relative to the specified frame

        Args:
            frame (str): The frame to which the jacobian will be computed

        Returns:
            J_dot (casADi function): The Jacobian derivative relative to the frame
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_velocity = cs.SX.sym("v_b", 6)
        joint_velocities = cs.SX.sym("s_dot", self.NDoF)
        J_dot = self.rbdalgos.jacobian_dot(
            frame, base_transform, joint_positions, base_velocity, joint_velocities
        )
        return cs.Function(
            "J_dot",
            [base_transform, joint_positions, base_velocity, joint_velocities],
            [J_dot.array],
            self.f_opts,
        )

    def CoM_position_fun(self) -> cs.Function:
        """Returns the CoM positon

        Returns:
            CoM (casADi function): The CoM position
        """
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_transform = cs.SX.sym("H", 4, 4)
        com_pos = self.rbdalgos.CoM_position(base_transform, joint_positions)
        return cs.Function(
            "CoM_pos", [base_transform, joint_positions], [com_pos.array], self.f_opts
        )

    def bias_force_fun(self) -> cs.Function:
        """Returns the bias force of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Returns:
            h (casADi function): the bias force
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_velocity = cs.SX.sym("v_b", 6)
        joint_velocities = cs.SX.sym("s_dot", self.NDoF)
        h = self.rbdalgos.rnea(
            base_transform, joint_positions, base_velocity, joint_velocities, self.g
        )
        return cs.Function(
            "h",
            [base_transform, joint_positions, base_velocity, joint_velocities],
            [h.array],
            self.f_opts,
        )

    def coriolis_term_fun(self) -> cs.Function:
        """Returns the coriolis term of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Returns:
            C (casADi function): the Coriolis term
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        base_velocity = cs.SX.sym("v_b", 6)
        joint_velocities = cs.SX.sym("s_dot", self.NDoF)
        # set in the bias force computation the gravity term to zero
        C = self.rbdalgos.rnea(
            base_transform,
            joint_positions,
            base_velocity,
            joint_velocities,
            np.zeros(6),
        )
        return cs.Function(
            "C",
            [base_transform, joint_positions, base_velocity, joint_velocities],
            [C.array],
            self.f_opts,
        )

    def gravity_term_fun(self) -> cs.Function:
        """Returns the gravity term of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Returns:
            G (casADi function): the gravity term
        """
        base_transform = cs.SX.sym("H", 4, 4)
        joint_positions = cs.SX.sym("s", self.NDoF)
        # set in the bias force computation the velocity to zero
        G = self.rbdalgos.rnea(
            base_transform, joint_positions, np.zeros(6), np.zeros(self.NDoF), self.g
        )
        return cs.Function(
            "G", [base_transform, joint_positions], [G.array], self.f_opts
        )

    def get_total_mass(self) -> float:
        """Returns the total mass of the robot

        Returns:
            mass: The total mass
        """
        return self.rbdalgos.get_total_mass()

    def mass_matrix(
        self, base_transform: Union[cs.SX, cs.DM], joint_positions: Union[cs.SX, cs.DM]
    ):
        """Returns the Mass Matrix functions computed the CRBA

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            M (jax): Mass Matrix
        """
        [M, _] = self.rbdalgos.crba(base_transform, joint_positions)
        return M.array

    def centroidal_momentum_matrix(
        self, base_transform: Union[cs.SX, cs.DM], joint_positions: Union[cs.SX, cs.DM]
    ):
        """Returns the Centroidal Momentum Matrix functions computed the CRBA

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            Jcc (Union[cs.SX, cs.DM]): Centroidal Momentum matrix
        """
        [_, Jcm] = self.rbdalgos.crba(base_transform, joint_positions)
        return Jcm.array

    def relative_jacobian(self, frame: str, joint_positions: Union[cs.SX, cs.DM]):
        """Returns the Jacobian between the root link and a specified frame frames

        Args:
            frame (str): The tip of the chain
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            J (Union[cs.SX, cs.DM]): The Jacobian between the root and the frame
        """
        return self.rbdalgos.relative_jacobian(frame, joint_positions).array

    def jacobian_dot(
        self,
        frame: str,
        base_transform: Union[cs.SX, cs.DM],
        joint_positions: Union[cs.SX, cs.DM],
        base_velocity: Union[cs.SX, cs.DM],
        joint_velocities: Union[cs.SX, cs.DM],
    ) -> Union[cs.SX, cs.DM]:
        """Returns the Jacobian derivative relative to the specified frame

        Args:
            frame (str): The frame to which the jacobian will be computed
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position
            base_velocity (Union[cs.SX, cs.DM]): The base velocity in mixed representation
            joint_velocities (Union[cs.SX, cs.DM]): The joint velocities

        Returns:
            Jdot (Union[cs.SX, cs.DM]): The Jacobian derivative relative to the frame
        """
        return self.rbdalgos.jacobian_dot(
            frame, base_transform, joint_positions, base_velocity, joint_velocities
        ).array

    def forward_kinematics(
        self,
        frame: str,
        base_transform: Union[cs.SX, cs.DM],
        joint_positions: Union[cs.SX, cs.DM],
    ):
        """Computes the forward kinematics relative to the specified frame

        Args:
            frame (str): The frame to which the fk will be computed
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            H (Union[cs.SX, cs.DM]): The fk represented as Homogenous transformation matrix
        """
        return self.rbdalgos.forward_kinematics(
            frame, base_transform, joint_positions
        ).array

    def jacobian(self, frame: str, base_transform, joint_positions):
        """Returns the Jacobian relative to the specified frame

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            s (Union[cs.SX, cs.DM]): The joints position
            frame (str): The frame to which the jacobian will be computed

        Returns:
            J_tot (Union[cs.SX, cs.DM]): The Jacobian relative to the frame
        """
        return self.rbdalgos.jacobian(frame, base_transform, joint_positions).array

    def bias_force(
        self,
        base_transform: Union[cs.SX, cs.DM],
        joint_positions: Union[cs.SX, cs.DM],
        base_velocity: Union[cs.SX, cs.DM],
        joint_velocities: Union[cs.SX, cs.DM],
    ) -> Union[cs.SX, cs.DM]:
        """Returns the bias force of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position
            base_velocity (Union[cs.SX, cs.DM]): The base velocity in mixed representation
            joint_velocities (Union[cs.SX, cs.DM]): The joints velocity

        Returns:
            h (Union[cs.SX, cs.DM]): the bias force
        """
        return self.rbdalgos.rnea(
            base_transform, joint_positions, base_velocity, joint_velocities, self.g
        ).array

    def coriolis_term(
        self,
        base_transform: Union[cs.SX, cs.DM],
        joint_positions: Union[cs.SX, cs.DM],
        base_velocity: Union[cs.SX, cs.DM],
        joint_velocities: Union[cs.SX, cs.DM],
    ) -> Union[cs.SX, cs.DM]:
        """Returns the coriolis term of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position
            base_velocity (Union[cs.SX, cs.DM]): The base velocity in mixed representation
            joint_velocities (Union[cs.SX, cs.DM]): The joints velocity

        Returns:
            C (Union[cs.SX, cs.DM]): the Coriolis term
        """
        return self.rbdalgos.rnea(
            base_transform,
            joint_positions,
            base_velocity.reshape(6, 1),
            joint_velocities,
            np.zeros(6),
        ).array

    def gravity_term(
        self, base_transform: Union[cs.SX, cs.DM], joint_positions: Union[cs.SX, cs.DM]
    ) -> Union[cs.SX, cs.DM]:
        """Returns the gravity term of the floating-base dynamics equation,
        using a reduced RNEA (no acceleration and external forces)

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            G (Union[cs.SX, cs.DM]): the gravity term
        """
        return self.rbdalgos.rnea(
            base_transform,
            joint_positions,
            np.zeros(6).reshape(6, 1),
            np.zeros(self.NDoF),
            self.g,
        ).array

    def CoM_position(
        self, base_transform: Union[cs.SX, cs.DM], joint_positions: Union[cs.SX, cs.DM]
    ) -> Union[cs.SX, cs.DM]:
        """Returns the CoM positon

        Args:
            base_transform (Union[cs.SX, cs.DM]): The homogenous transform from base to world frame
            joint_positions (Union[cs.SX, cs.DM]): The joints position

        Returns:
            CoM (Union[cs.SX, cs.DM]): The CoM position
        """
        return self.rbdalgos.CoM_position(base_transform, joint_positions).array
